"""
Rank command.

The report command gives a table of files sorted according their ranking scheme
of a specified metric.
Will compare the values between files and return a sorted table.

TODO: Layer on Click invocation in operators section, __main__.py file
"""
import os

import tabulate
import operator as op
from pathlib import Path

from wily import logger, format_revision, format_date
from wily.archivers import resolve_archiver
from wily.config import DEFAULT_GRID_STYLE, DEFAULT_PATH
from wily.state import State
from wily.operators import resolve_metric_as_tuple, MetricType

import radon.cli.harvest


def rank(config, path, metric, revision_index, limit, threshold, descending):
    """
    Rank command ordering files, methods or functions using metrics.

    :param config: The configuration
    :type config: :class:'wily.config.WilyConfig'

    :param path: The path to the file
    :type path ''str''

    :param metric: Name of the metric to report on
    :type metric: ''str''

    :param revision_index: Version of git repository to revert to.
    :type revision_index: ``str``

    :param limit: Limit the number of items in the table
    :type  limit: ``int``

    :param threshold: For total values beneath the threshold return a non-zero exit code
    :type  threshold: ``int``

    :return: Sorted table of all files in path, sorted in order of metric.
    """
    logger.debug("Running rank command")

    data = []

    operator, metric = resolve_metric_as_tuple(metric)
    operator = operator.name

    state = State(config)

    if not revision_index:
        target_revision = state.index[state.default_archiver].last_revision
    else:
        rev = resolve_archiver(state.default_archiver).cls(config).find(revision_index)
        logger.debug(f"Resolved {revision_index} to {rev.key} ({rev.message})")
        try:
            target_revision = state.index[state.default_archiver][rev.key]
        except KeyError:
            logger.error(
                f"Revision {revision_index} is not in the cache, make sure you have run wily build."
            )
            exit(1)

    logger.info(
        f"-----------Rank for {metric.description} for {format_revision(target_revision.revision.key)} by {target_revision.revision.author_name} on {format_date(target_revision.revision.date)}.------------"
    )

    if path is None:
        files = target_revision.get_paths(config, state.default_archiver, operator)
        logger.debug(f"Analysing {files}")
    else:
        # Resolve target paths when the cli has specified --path
        if config.path != DEFAULT_PATH:
            targets = [str(Path(config.path) / Path(path))]
        else:
            targets = [path]

        # Expand directories to paths
        files = [
            os.path.relpath(fn, config.path)
            for fn in radon.cli.harvest.iter_filenames(targets)
        ]
        logger.debug(f"Targeting - {files}")

    for item in files:
        for archiver in state.archivers:
            try:
                logger.debug(
                    f"Fetching metric {metric.name} for {operator} in {str(item)}"
                )
                val = target_revision.get(
                    config, archiver, operator, str(item), metric.name
                )
                value = val
                data.append((item, value))
            except KeyError:
                logger.debug(f"Could not find file {item} in index")

    # Sort by ideal value
    data = sorted(data, key=op.itemgetter(1), reverse=descending)

    if limit:
        data = data[:limit]

    # Tack on the total row at the end
    total = metric.aggregate(rev[1] for rev in data)
    data.append(["Total", total])

    headers = ("File", metric.description)
    print(
        tabulate.tabulate(
            headers=headers, tabular_data=data, tablefmt=DEFAULT_GRID_STYLE
        )
    )

    if threshold and total < threshold:
        logger.error(
            f"Total value below the specified threshold: {total} < {threshold}"
        )
        exit(1)



from click.testing import CliRunner

import wily.__main__ as main
from git import Actor, Repo


def test_rank_no_cache(tmpdir):
    """ Test the rank feature with no cache """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", tmpdir, "rank", "src/test.py"])
    assert result.exit_code == 1, result.stdout


def test_rank_single_file_default_metric(builddir):
    """ Test the rank feature with default (AimLow) metric on a single file """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank", "src/test.py"])
    assert result.exit_code == 0, result.stdout


def test_rank_directory_default_metric(builddir):
    """ Test the rank feature with default (AimLow) metric on a directory """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank", "src/"])
    assert result.exit_code == 0, result.stdout


def test_rank_directory_default_metric_no_path(builddir):
    """ Test the rank feature with no path """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank"])
    assert result.exit_code == 0, result.stdout


def test_rank_directory_default_metric_master(builddir):
    """ Test the rank feature with a specific revision. """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank", "-r", "master"])
    assert result.exit_code == 0, result.stdout


def test_rank_directory_default_invalid_revision(builddir):
    """ Test the rank feature with an invalid revision. """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank", "-r", "xyz"])
    assert result.exit_code == 1, result.stdout


def test_rank_directory_default_unindexed_revision(builddir):
    """ Test the rank feature with an unindexed revision. """
    repo = Repo(builddir)
    with open(builddir / "test.py", "w") as test_txt:
        test_txt.write("import abc")

    index = repo.index
    index.add([str(builddir / "test.py")])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    commit = index.commit(
        "unindexed commit",
        author=author,
        committer=committer,
        author_date="Thu, 28 Apr 2019 22:13:13 +0200",
        commit_date="Thu, 28 Apr 2019 22:13:13 +0200",
    )
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank", "-r", commit.hexsha])
    assert result.exit_code == 1, result.stdout


def test_rank_single_file_informational(builddir):
    """ Test the rank feature with Informational metric """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "rank", "src/test.py", "raw.loc"]
    )
    assert result.exit_code == 0, result.stdout


def test_rank_directory_custom_metric(builddir):
    """ Test the rank feature with AimHigh metric """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "rank", "src/", "raw.comments"]
    )
    assert result.exit_code == 0, result.stdout


def test_rank_directory_no_path_target(builddir):
    """ Test the rank feature with no path target """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["rank", "src/", "raw.comments"])
    assert result.exit_code == 0, result.stdout


def test_rank_directory_limit(builddir):
    """ Test the rank feature with limit """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "rank", "src/", "raw.comments", "-l 2"]
    )
    assert result.exit_code == 0, result.stdout


def test_rank_directory_desc(builddir):
    """ Test the rank feature descending order """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "rank", "src/", "raw.comments", "--desc"]
    )
    assert result.exit_code == 0, result.stdout


def test_rank_directory_invalid_key(builddir):
    """ Test the rank feature descending order with an invalid key """
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "rank", "invalid/", "raw.comments", "--desc"]
    )
    assert result.exit_code == 0, result.stdout


def test_rank_directory_asc(builddir):
    """ Test the rank feature ascending order"""
    runner = CliRunner()
    result = runner.invoke(
        main.cli, ["--path", builddir, "rank", "src/", "raw.comments", "--asc"]
    )
    assert result.exit_code == 0, result.stdout


def test_rank_total_above_threshold(builddir):
    """ Test the rank feature with total above threshold """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank", "--threshold=20"])
    assert result.exit_code == 0, result.stdout


def test_rank_total_below_threshold(builddir):
    """ Test the rank feature with total below threshold """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", builddir, "rank", "--threshold=100"])
    assert result.exit_code == 1, result.stdout



"""
Integration tests that will create a repository with multiple files
and test the skipping of unchanged files does not impact the index.
"""
import sys
import pathlib
import json
from click.testing import CliRunner
from git import Repo, Actor

import wily.__main__ as main

_path1 = "src\\test1.py" if sys.platform == "win32" else "src/test1.py"
_path2 = "src\\test2.py" if sys.platform == "win32" else "src/test2.py"


def test_skip_files(tmpdir, cache_path):
    """
    Test that files which were not changed are still added to each index
    """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir) / "src"
    tmppath.mkdir()

    # Write two test files to the repo
    with open(tmppath / "test1.py", "w") as test1_txt:
        test1_txt.write("import abc")

    with open(tmppath / "test2.py", "w") as test2_txt:
        test2_txt.write("import cde")

    index = repo.index
    index.add([str(tmppath / "test1.py"), str(tmppath / "test2.py")])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    commit = index.commit("commit two files", author=author, committer=committer)

    # Change the second file and commit that
    with open(tmppath / "test2.py", "w") as test2_txt:
        test2_txt.write("import zzz\nprint(1)")

    repo.index.add([str(tmppath / "test2.py")])
    commit2 = repo.index.commit(
        "commit the second file only", author=author, committer=committer
    )

    # Change the first file and commit that
    with open(tmppath / "test1.py", "w") as test2_txt:
        test2_txt.write("import zzz\nprint(1)")

    repo.index.add([str(tmppath / "test1.py")])
    commit3 = repo.index.commit(
        "commit the first file only", author=author, committer=committer
    )

    repo.close()

    # build the wily cache and test its contents
    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--debug", "--path", tmpdir, "--cache", cache_path, "build", str(tmppath)],
    )
    assert result.exit_code == 0, result.stdout

    # Check that the index files were created
    cache_path = pathlib.Path(cache_path)
    assert cache_path.exists()
    index_path = cache_path / "git" / "index.json"
    assert index_path.exists()
    rev_path = cache_path / "git" / (commit.name_rev.split(" ")[0] + ".json")
    assert rev_path.exists()

    # Inspect the contents of the index for the existence of both files
    with open(index_path) as index_file:
        index = json.load(index_file)

    assert len(index) == 3

    # Look at the first commit
    with open(rev_path) as rev_file:
        data = json.load(rev_file)

    assert "raw" in data["operator_data"]
    assert _path1 in data["operator_data"]["raw"]
    assert _path2 in data["operator_data"]["raw"]

    # Look at the second commit
    rev2_path = cache_path / "git" / (commit2.name_rev.split(" ")[0] + ".json")
    assert rev2_path.exists()

    with open(rev2_path) as rev2_file:
        data2 = json.load(rev2_file)

    assert "raw" in data2["operator_data"]
    assert _path1 in data2["operator_data"]["raw"]
    assert _path2 in data2["operator_data"]["raw"]

    # Look at the third commit
    rev3_path = cache_path / "git" / (commit3.name_rev.split(" ")[0] + ".json")
    assert rev3_path.exists()

    with open(rev3_path) as rev3_file:
        data3 = json.load(rev3_file)

    assert "raw" in data3["operator_data"]
    assert _path1 in data3["operator_data"]["raw"]
    assert _path2 in data3["operator_data"]["raw"]



"""Language/i18n support for the CLI."""
# -*- coding: UTF-8 -*-

import gettext
import os

localedir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'locales')
trans = gettext.translation('messages', localedir=localedir, fallback=True)
trans.install()
_ = trans.gettext


"""
Raw statistics operator.

Includes insights like lines-of-code, number of comments. Does not measure complexity.
"""
import radon.cli.harvest as harvesters
from radon.cli import Config

from wily import logger
from wily.operators import BaseOperator, MetricType, Metric


class RawMetricsOperator(BaseOperator):
    """Raw Metrics Operator."""

    name = "raw"
    defaults = {
        "exclude": None,
        "ignore": None,
        "summary": False,
        "include_ipynb": True,
        "ipynb_cells": True,
    }
    metrics = (
        Metric("loc", _("Lines of Code"), int, MetricType.Informational, sum),
        Metric("lloc", _("L Lines of Code"), int, MetricType.AimLow, sum),
        Metric("sloc", _("S Lines of Code"), int, MetricType.AimLow, sum),
        Metric("comments", _("Multi-line comments"), int, MetricType.AimHigh, sum),
        Metric("multi", _("Multi lines"), int, MetricType.Informational, sum),
        Metric("blank", _("blank lines"), int, MetricType.Informational, sum),
        Metric(
            "single_comments",
            _("Single comment lines"),
            int,
            MetricType.Informational,
            sum,
        ),
    )
    default_metric_index = 0  # LOC

    def __init__(self, config, targets):
        """
        Instantiate a new raw operator.

        :param config: The wily configuration.
        :type  config: :class:`WilyConfig`
        """
        # TODO: Use config from wily.cfg for harvester
        logger.debug(f"Using {targets} with {self.defaults} for Raw metrics")
        self.harvester = harvesters.RawHarvester(
            targets, config=Config(**self.defaults)
        )

    def run(self, module, options):
        """
        Run the operator.

        :param module: The target module path.
        :type  module: ``str``

        :param options: Any runtime options.
        :type  options: ``dict``

        :return: The operator results.
        :rtype: ``dict``
        """
        logger.debug("Running raw harvester")
        results = {}
        for filename, metrics in dict(self.harvester.results).items():
            results[filename] = {"total": metrics}
        return results



import sys
from click.testing import CliRunner

import wily.__main__ as main

_path = "src\\test.ipynb" if sys.platform == "win32" else "src/test.ipynb"


def test_index_with_ipynb(ipynbbuilddir):
    """
    Test that index works with a build
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", ipynbbuilddir, "index"])
    assert result.stdout.count("An author") >= 3
    assert result.exit_code == 0, result.stdout


def test_index_and_message_with_ipynb(ipynbbuilddir):
    """
    Test that index works with a build
    """
    runner = CliRunner()
    result = runner.invoke(main.cli, ["--path", ipynbbuilddir, "index", "--message"])
    assert result.stdout.count("An author") >= 3
    assert "empty notebook" in result.stdout
    assert "single cell" in result.stdout
    assert "second cell" in result.stdout
    assert result.exit_code == 0, result.stdout


def test_report(ipynbbuilddir):
    """
    Test that report works with a build and a specific metric
    """
    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--path", ipynbbuilddir, "report", _path, "raw.multi", "maintainability.rank"],
    )
    assert result.exit_code == 0, result.stdout
    assert "Not found" not in result.stdout



import pathlib
from textwrap import dedent
import shutil
import tempfile
from time import time

import pytest
from click.testing import CliRunner
from git import Repo, Actor

import wily.__main__ as main


@pytest.fixture
def gitdir(tmpdir):
    """ Create a project and add code to it """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir)
    testpath = tmppath / "src" / "test.py"
    (tmppath / "src").mkdir()
    # Write a test file to the repo
    with open(testpath, "w") as test_txt:
        test_txt.write("import abc")

    index = repo.index
    index.add([str(testpath)])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    index.commit(
        "basic test",
        author=author,
        committer=committer,
        author_date="Thu, 07 Apr 2019 22:13:13 +0200",
        commit_date="Thu, 07 Apr 2019 22:13:13 +0200",
    )

    first_test = """
    import abc
    foo = 1
    def function1():
        a = 1 + 1
    
    class Class1(object):
        def method(self):
            b = 1 + 5
    """
    with open(testpath, "w") as test_txt:
        test_txt.write(dedent(first_test))

    index.add([str(testpath)])
    index.commit(
        "add line",
        author=author,
        committer=committer,
        author_date="Mon, 10 Apr 2019 22:13:13 +0200",
        commit_date="Mon, 10 Apr 2019 22:13:13 +0200",
    )

    second_test = """
    import abc
    foo = 1
    def function1():
        a = 1 + 1
    class Class1(object):
        def method(self):
            b = 1 + 5
            if b == 6:
                return 'banana'
    """

    with open(testpath, "w") as test_txt:
        test_txt.write(dedent(second_test))

    index.add([str(testpath)])
    index.commit(
        "remove line",
        author=author,
        committer=committer,
        author_date="Thu, 14 Apr 2019 22:13:13 +0200",
        commit_date="Thu, 14 Apr 2019 22:13:13 +0200",
    )

    yield tmpdir
    repo.close()


@pytest.fixture
def builddir(gitdir):
    """
    A directory with a wily cache
    """
    tmppath = pathlib.Path(gitdir)
    runner = CliRunner()
    result1 = runner.invoke(
        main.cli, ["--debug", "--path", gitdir, "build", str(tmppath / "src")]
    )
    assert result1.exit_code == 0, result1.stdout

    result2 = runner.invoke(main.cli, ["--debug", "--path", gitdir, "index"])
    assert result2.exit_code == 0, result2.stdout

    yield gitdir

    result1 = runner.invoke(main.cli, ["--debug", "--path", gitdir, "clean", "-y"])
    assert result1.exit_code == 0, result1.stdout


@pytest.fixture
def ipynbgitdir(tmpdir):
    """
    A fixture that provides a directory and a working Git DB,
    contains a single IPython notebook that has 3 revisions.
    """
    _NB_FOOTER = """
     "metadata": {
          "kernelspec": {
           "display_name": "Python 3",
           "language": "python",
           "name": "python3"
          },
          "language_info": {
           "codemirror_mode": {
            "name": "ipython",
            "version": 3
           },
           "file_extension": ".py",
           "mimetype": "text/x-python",
           "name": "python",
           "nbconvert_exporter": "python",
           "pygments_lexer": "ipython3",
           "version": "3.4.2"
          }
         },
     "nbformat": 4,
     "nbformat_minor": 0
    """
    repo = Repo.init(path=tmpdir)
    tmppath = pathlib.Path(tmpdir)
    testpath = tmppath / "src" / "test.ipynb"
    (tmppath / "src").mkdir()
    # Write a test file to the repo
    with open(testpath, "w") as test_txt:
        test_txt.write('{"cells": [],' + _NB_FOOTER + "}")

    index = repo.index
    index.add([str(testpath)])

    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    index.commit("empty notebook", author=author, committer=committer)

    first_test = (
        """{
     "cells": [
      {
       "cell_type": "code",
       "metadata": {},
       "language": "python",
       "source": [
        "import abc\\n",
        "foo = 1\\n",
        "def function1():\\n",
        "    a = 1 + 1\\n",
        "\\n",
        "class Class1(object):\\n",
        "    def method(self):\\n",
        "        b = 1 + 5\\n"
       ],
       "outputs": [],
       "execution_count": 0,
       "input": []
      }
      ],
    """
        + _NB_FOOTER
        + "}"
    )
    with open(testpath, "w") as test_txt:
        test_txt.write(dedent(first_test))

    index.add([str(testpath)])
    index.commit("single cell", author=author, committer=committer)

    second_test = (
        """{
     "cells": [
      {
       "cell_type": "code",
       "metadata": {},
       "language": "python",
       "source": [
        "import abc\\n",
        "foo = 1\\n",
        "def function1():\\n",
        "    a = 1 + 1\\n",
        "class Class1(object):\\n",
        "    def method(self):\\n",
        "        b = 1 + 5\\n",
        "        if b == 6:\\n",
        "            return 'banana'\\n"
       ],
       "outputs": [],
       "execution_count": 0,
       "input": []
      },
      {
       "cell_type": "code",
       "metadata": {},
       "language": "python",
       "source": [
        "foo = 1\\n",
        "class Class1(object):\\n",
        "    def method(self):\\n",
        "        b = 1 + 5\\n",
        "        if b == 6:\\n",
        "            return 'banana'\\n"
       ],
       "outputs": [],
       "execution_count": 0,
       "input": []
      }
      ],
    """
        + _NB_FOOTER
        + "}"
    )

    with open(testpath, "w") as test_txt:
        test_txt.write(dedent(second_test))

    index.add([str(testpath)])
    index.commit("second cell", author=author, committer=committer)

    yield tmpdir
    repo.close()


@pytest.fixture
def ipynbbuilddir(ipynbgitdir):
    """
    The ipynbgitdir fixture converted into a wily cache index.
    """
    tmppath = pathlib.Path(ipynbgitdir)

    config = """
    [wily]
    include_ipynb = true
    ipynb_cells = true
    """
    config_path = tmppath / "wily.cfg"
    with open(config_path, "w") as config_f:
        config_f.write(config)

    runner = CliRunner()
    result1 = runner.invoke(
        main.cli, ["--debug", "--path", ipynbgitdir, "build", str(tmppath / "src")]
    )
    assert result1.exit_code == 0, result1.stdout

    result2 = runner.invoke(main.cli, ["--debug", "--path", ipynbgitdir, "index"])
    assert result2.exit_code == 0, result2.stdout

    yield ipynbgitdir

    result1 = runner.invoke(main.cli, ["--debug", "--path", ipynbgitdir, "clean", "-y"])
    assert result1.exit_code == 0, result1.stdout


@pytest.fixture(autouse=True)
def cache_path(monkeypatch):
    """
    Configure wily cache and home path, clean up cache afterward
    """
    tmp = tempfile.mkdtemp()
    monkeypatch.setenv("HOME", tmp)
    yield tmp
    shutil.rmtree(tmp)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wily', 'wily.archivers', 'wily.commands', 'wily.helper', 'wily.operators']

package_data = \
{'': ['*'],
 'wily': ['locales/*',
          'locales/de/*',
          'locales/de/LC_MESSAGES/*',
          'locales/en/*',
          'locales/en/LC_MESSAGES/*',
          'locales/en_AU/*',
          'locales/en_AU/LC_MESSAGES/*',
          'locales/ja/*',
          'locales/ja/LC_MESSAGES/*',
          'templates/*',
          'templates/css/*']}

package_dir = \
{'': 'src'}

install_requires = \
['gitpython>=3.0.0,<4.0.0',
 'radon>=4.0.0,<4.1.0',
 'click>=7.0,<8.0',
 'nbformat>=4.4.0,<4.5.0',
 'colorlog>=4.0.0,<5.0.0',
 'tabulate>=0.8.2,<1.0.0',
 'plotly>=4.0.0,<5.0.0',
 'progress==1.4']

extras_require = \
{":python_version == '3.6'": ['dataclasses']}

entry_points = \
{'console_scripts': ['wily = wily.__main__:cli']}

setup(name='wily',
      version='1.19.0',
      description='A Python application for tracking, reporting on timing and complexity in tests and applications.',
      author='Anthony Shaw',
      author_email='anthonyshaw@apache.org',
      url='https://github.com/tonybaloney/wily',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.6',
     )



"""
Parameterised tests for each operator (and some combinations).

Build them and test out some of the metrics/commands work correctly.
"""
import pytest
import sys
from click.testing import CliRunner
import pathlib
from textwrap import dedent

from git import Repo, Actor

import wily.__main__ as main

_path = "src\\test.py" if sys.platform == "win32" else "src/test.py"

operators = (
    "halstead",
    "cyclomatic",
    "maintainability",
    "raw",
    "halstead,cyclomatic",
    "maintainability,raw",
    "halstead,cyclomatic,maintainability,raw",
)


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@pytest.mark.parametrize("operator", operators)
def test_operator(operator, gitdir):
    runner = CliRunner()
    result = runner.invoke(
        main.cli,
        ["--debug", "--path", gitdir, "build", "src", "-o", operator],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stdout

    result = runner.invoke(
        main.cli, ["--debug", "--path", gitdir, "report", _path], catch_exceptions=False
    )
    assert result.exit_code == 0, result.stdout

    result = runner.invoke(
        main.cli,
        ["--debug", "--path", gitdir, "diff", _path, "--all"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout

    complex_test = """
            import abc
            foo = 1
            def function1():
                a = 1 + 1
                if a == 2:
                    print(1)
            class Class1(object):
                def method(self):
                    b = 1 + 5
                    if b == 6:
                        if 1==2:
                           if 2==3:
                              print(1)
            """

    with open(pathlib.Path(gitdir) / "src" / "test.py", "w") as test_py:
        test_py.write(dedent(complex_test))

    result = runner.invoke(
        main.cli,
        ["--debug", "--path", gitdir, "diff", _path, "--all"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stdout
    assert "test.py" in result.stdout


@pytest.mark.parametrize("operator", operators)
def test_operator_on_code_with_metric_named_objects(operator, tmpdir):
    code_with_metric_named_objects = """

    # CyclomaticComplexity
    def complexity(): pass

    # Halstead
    def h1(): pass
    def h2(): pass
    def N1(): pass
    def N2(): pass
    def vocabulary(): pass
    def length(): pass
    def volume(): pass
    def difficulty(): pass
    def error(): pass

    # Maintainability
    def rank(): pass
    def mi(): pass

    # RawMetrics
    def loc(): pass
    def lloc(): pass
    def sloc(): pass
    def comments(): pass
    def multi(): pass
    def blank(): pass
    def single_comments(): pass

    """

    testpath = pathlib.Path(tmpdir) / "test.py"
    author = Actor("An author", "author@example.com")
    committer = Actor("A committer", "committer@example.com")

    with open(testpath, "w") as test_py:
        test_py.write(dedent(code_with_metric_named_objects))

    with Repo.init(path=tmpdir) as repo:
        repo.index.add(["test.py"])
        repo.index.commit("add test.py", author=author, committer=committer)

    runner = CliRunner()

    result = runner.invoke(
        main.cli,
        ["--debug", "--path", tmpdir, "build", str(testpath), "-o", operator],
        catch_exceptions=False,
    )
    assert result.exit_code == 0, result.stdout



"""
A Python application for tracking, reporting on timing and complexity in tests and applications.

           B        BB*
         BBBB      BBB
       BBBBBB%$%*SBBB&
      BBBBS&@$@$SBBBB
     BBS@$$$$@@$#BBB. .
    BB&$$@@@@@@$$@#! !..
    B&$*$@@@*.@@@$$@&&@&$             BB
    %$$!$@$@$$@@@@@@@@@$$B       SBBBBS###B
    %$@@$$$$@@$$$@&$@@@$$    SBBBBBBBBSSSSSB
     %$@#BBBBBBS&&@@@@@$$$%#BBBBBBBBBBBBBBB
  *@BBBBBBBBBBBBS@$@@@@@@@%SBBBBBBBBBBBBBB
@$$@BBBBBBBBBBB#@$@@@@@@@@$#BBBBBBBBBB                  BBBBBBBBBBB    BBB    BBBBBBBBBBBB   BBBBBBB
@$$$SBBBBSSS&!  *@@@@@@@@@$$#BBBBB#                     BBBBBBBBBBB    BBB    BBBBBBBBBBBB   BBBBBBB
@$@$$@&&$@&%S   $@@@@@@@@@@@$$#                            BBBBBB     BBBBB      BBB  BBBB     BBBBB
@$$$$$$$$&BBB. $#&&@@@@@@@@@$                               BBBBB    BBBBBBB    BBB     BB     BBBBB BBB   BBB  BB  BB
  %%$$@@$$@@*!@@*!*$@@@@@$$*                                 BBBBB   BBBBBBB    BBB BBBBBB     BBBBB BBBBBBBBB  BBBBBBB
     $$$$$$$@&*     .$@@@$$                                  BBBBB  BBB BBBBB  BBB  BBBBBB     BBBBB   BBBBB     BBBB
         %$$@@        $@@$$                                   BBBBB BBB BBBBB  BBB    BBBB     BBBBB    BBBBB    BBB
           $$@*        @@$$                                   BBBBB BB   BBBBBBBB     BBBB     BBBBB     BBBBB  BBB
            %@$        %@@$%                                   BBBBBBB    BBBBBB      BBBB     BBBBB      BBBBBBBB
             @!        %@@$$@$$$$$$$$                           BBBBB     BBBBBB      BBBB     BBBBB       BBBBBB   BBB
            $@!       !@@@@@@@@@@@@@@$$$$$$                     BBBBB      BBBB     BBBBBBBB BBBBBBBBB      BBBBB  BBBBB
           $$@@*.   !$&@@@@@@@@@@@@@@@@@@@@@@S                   BBB        BBB     BBBBBBBB BBBBBBBBB      BBBB   BBBBB
           $$@@@@@@@&@@@@@@@@@@@@@@@@@@&&&&@!!$@                                                           BBBB
         *$$@@@@@@@@@@@@@@@@@@@@@@@@@@@$ .*@!  %                                                          BBBB
        @$$@@@@@@@$*@@@@@@@@$$$@@@@@@@&&%   .      .                                                  BB BBBB
        %$@@$$@@@%   $@@@$$$$$$$@@@@@$*%@@*                                                          BBBBBBB
        $$$$$@@$      $@@#BBBBBB@@@&&@%!.........                                                    BBBBBB
        *&SBBBBBB       $BBBBBBB#@&@$$$%%*!........
         BBBBBBBB*       #BBBBBBB*
          BBBBBBBB        BBBBBBBB
      %$%%BBBBBBS%$$$$$$$$%&BBBBBBS@$$$$$%%$$$$$$%$$$&B*%
      %%$%$&&S#&$$@@@@@@@@@$$@&&&@$@@@@@@@@@@@@@@@@$$%%$$
               .%%$$$$%%%$%%%%***%%%%%$%%$$$$$%%$%*&
"""
import tempfile
import colorlog
import logging
import datetime

__version__ = "1.19.0"

_, WILY_LOG_NAME = tempfile.mkstemp(suffix="wily_log")

_handler = colorlog.StreamHandler()
_handler.setFormatter(colorlog.ColoredFormatter("%(log_color)s%(message)s"))

_filehandler = logging.FileHandler(WILY_LOG_NAME, mode="w+")
_filehandler.setLevel(logging.DEBUG)

logger = colorlog.getLogger(__name__)
logger.addHandler(_handler)
logger.addHandler(_filehandler)

""" Max number of characters of the Git commit to print """
MAX_MESSAGE_WIDTH = 50


def format_date(timestamp):
    """Reusable timestamp -> date."""
    return datetime.date.fromtimestamp(timestamp).isoformat()


def format_datetime(timestamp):
    """Reusable timestamp -> datetime."""
    return datetime.datetime.fromtimestamp(timestamp).isoformat()


def format_revision(sha):
    """Return a shorter git sha."""
    return sha[:7]



"""
Builds a cache based on a source-control history.

TODO : Convert .gitignore to radon ignore patterns to make the build more efficient.

"""
import os
import pathlib
import multiprocessing
from progress.bar import Bar

from wily import logger
from wily.state import State

from wily.archivers.git import InvalidGitRepositoryError
from wily.archivers import FilesystemArchiver

from wily.operators import resolve_operator


def run_operator(operator, revision, config, targets):
    """
    Run an operator for the multiprocessing pool.

    :param operator: The operator name
    :type  operator: ``str``

    :param revision: The revision index
    :type  revision: :class:`Revision`

    :param config: The runtime configuration
    :type  config: :class:`WilyConfig`

    :param targets: Files/paths to scan
    :type  targets: ``list`` of ``str``

    :rtype: ``tuple``
    :returns: A tuple of operator name (``str``), and data (``dict``)
    """
    instance = operator.cls(config, targets)
    logger.debug(f"Running {operator.name} operator on {revision}")

    data = instance.run(revision, config)

    # Normalize paths for non-seed passes
    for key in list(data.keys()):
        if os.path.isabs(key):
            rel = os.path.relpath(key, config.path)
            data[rel] = data[key]
            del data[key]

    return operator.name, data


def build(config, archiver, operators):
    """
    Build the history given a archiver and collection of operators.

    :param config: The wily configuration
    :type  config: :namedtuple:`wily.config.WilyConfig`

    :param archiver: The archiver to use
    :type  archiver: :namedtuple:`wily.archivers.Archiver`

    :param operators: The list of operators to execute
    :type operators: `list` of :namedtuple:`wily.operators.Operator`
    """
    try:
        logger.debug(f"Using {archiver.name} archiver module")
        archiver = archiver.cls(config)
        revisions = archiver.revisions(config.path, config.max_revisions)
    except InvalidGitRepositoryError:
        # TODO: This logic shouldn't really be here (SoC)
        logger.info(f"Defaulting back to the filesystem archiver, not a valid git repo")
        archiver = FilesystemArchiver(config)
        revisions = archiver.revisions(config.path, config.max_revisions)
    except Exception as e:
        if hasattr(e, "message"):
            logger.error(f"Failed to setup archiver: '{e.message}'")
        else:
            logger.error(f"Failed to setup archiver: '{type(e)} - {e}'")
        exit(1)

    state = State(config, archiver=archiver)
    # Check for existence of cache, else provision
    state.ensure_exists()

    index = state.index[archiver.name]

    # remove existing revisions from the list
    revisions = [revision for revision in revisions if revision not in index][::-1]

    logger.info(
        f"Found {len(revisions)} revisions from '{archiver.name}' archiver in '{config.path}'."
    )

    _op_desc = ",".join([operator.name for operator in operators])
    logger.info(f"Running operators - {_op_desc}")

    bar = Bar("Processing", max=len(revisions) * len(operators))
    state.operators = operators

    # Index all files the first time, only scan changes afterward
    seed = True
    prev_roots = None
    try:
        with multiprocessing.Pool(processes=len(operators)) as pool:
            for revision in revisions:
                # Checkout target revision
                archiver.checkout(revision, config.checkout_options)
                stats = {"operator_data": {}}

                if seed:
                    targets = config.targets
                else:  # Only target changed files
                    # TODO : Check that changed files are children of the targets
                    targets = [
                        str(pathlib.Path(config.path) / pathlib.Path(file))
                        for file in revision.files
                        # if any([True for target in config.targets if
                        #         target in pathlib.Path(pathlib.Path(config.path) / pathlib.Path(file)).parents])
                    ]

                # Run each operator as a separate process
                data = pool.starmap(
                    run_operator,
                    [(operator, revision, config, targets) for operator in operators],
                )
                # data is a list of tuples, where for each operator, it is a tuple of length 2,
                operator_data_len = 2
                # second element in the tuple, i.e data[i][1]) has the collected data
                for i in range(0, len(operators)):
                    if (
                        i < len(data)
                        and len(data[i]) >= operator_data_len
                        and len(data[i][1]) == 0
                    ):
                        logger.warning(
                            f"In revision {revision.key}, for operator {operators[i].name}: No data collected"
                        )

                # Map the data back into a dictionary
                for operator_name, result in data:
                    # find all unique directories in the results
                    roots = {pathlib.Path(entry).parents[0] for entry in result.keys()}
                    indices = set(result.keys())

                    # For a seed run, there is no previous change set, so use current
                    if seed:
                        prev_roots = roots
                        prev_indices = indices
                    roots = prev_roots | roots

                    # Copy the ir from any unchanged files from the prev revision
                    if not seed:
                        missing_indices = prev_indices - indices
                        # TODO: Check existence of file path.
                        for missing in missing_indices:
                            # Don't copy aggregate keys as their values may have changed
                            if missing in roots:
                                continue
                            # previous index may not have that operator
                            if operator_name not in prev_stats["operator_data"]:
                                continue
                            # previous index may not have file either
                            if (
                                missing
                                not in prev_stats["operator_data"][operator_name]
                            ):
                                continue
                            result[missing] = prev_stats["operator_data"][
                                operator_name
                            ][missing]

                    # Aggregate metrics across all root paths using the aggregate function in the metric
                    for root in roots:
                        # find all matching entries recursively
                        aggregates = [
                            path
                            for path in result.keys()
                            if root in pathlib.Path(path).parents
                        ]
                        result[str(root)] = {"total": {}}
                        # aggregate values
                        for metric in resolve_operator(operator_name).cls.metrics:
                            func = metric.aggregate
                            values = [
                                result[aggregate]["total"][metric.name]
                                for aggregate in aggregates
                                if aggregate in result
                                and metric.name in result[aggregate]["total"]
                            ]
                            if len(values) > 0:
                                result[str(root)]["total"][metric.name] = func(values)

                    prev_indices = set(result.keys())
                    prev_roots = roots
                    stats["operator_data"][operator_name] = result
                    bar.next()

                prev_stats = stats
                seed = False
                ir = index.add(revision, operators=operators)
                ir.store(config, archiver, stats)
        index.save()
        bar.finish()
    except Exception as e:
        logger.error(f"Failed to build cache: {type(e)}: '{e}'")
        raise e
    finally:
        # Reset the archive after every run back to the head of the branch
        archiver.finish()
